# passerelle - uniform access to multiple data sources and services
# Copyright (C) 2020  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.db import models
from django.shortcuts import get_object_or_404
from django.template import Context, Template
from django.core.urlresolvers import reverse
from django.utils.encoding import force_text
from django.utils.six.moves.urllib import parse as urlparse
from django.utils.translation import ugettext_lazy as _

from passerelle.utils.templates import validate_template
from passerelle.base.models import BaseResource, BaseQuery
from passerelle.utils.api import endpoint


class OpenDataSoft(BaseResource):
    service_url = models.CharField(
        _('Service URL'),
        max_length=256, blank=False,
        help_text=_('OpenDataSoft webservice URL'),
    )
    api_key = models.CharField(
        _('API key'),
        max_length=128, blank=True,
        help_text=_('API key used as credentials'),
    )

    category = _('Data Sources')

    class Meta:
        verbose_name = _('OpenDataSoft Web Service')

    def export_json(self):
        data = super(OpenDataSoft, self).export_json()
        data['queries'] = [query.export_json() for query in self.queries.all()]
        return data

    @classmethod
    def import_json_real(cls, overwrite, instance, data, **kwargs):
        data_queries = data.pop('queries', [])
        instance = super(OpenDataSoft, cls).import_json_real(overwrite, instance, data, **kwargs)
        queries = []
        if instance and overwrite:
            Query.objects.filter(resource=instance).delete()
        for data_query in data_queries:
            query = Query.import_json(data_query)
            query.resource = instance
            queries.append(query)
        Query.objects.bulk_create(queries)
        return instance

    @endpoint(
        perm='can_access',
        description=_('Search'),
        parameters={
            'dataset': {'description': _('Dataset')},
            'text_template': {'description': _('Text template')},
            'id': {'description': _('Record identifier')},
            'q': {'description': _('Full text query')},
            'limit': {'description': _('Maximum items')},
        })
    def search(self, request, dataset=None, text_template='', id=None, q=None, limit=None, **kwargs):
        scheme, netloc, path, params, query, fragment = urlparse.urlparse(self.service_url)
        path = urlparse.urljoin(path, 'api/records/1.0/search/')
        url = urlparse.urlunparse((scheme, netloc, path, params, query, fragment))

        if id is not None:
            query = 'recordid:%s' % id
        else:
            query = q
        params = {
            'dataset': dataset,
            'q': query,
        }
        if self.api_key:
            params.update({'apikey': self.api_key})
        if limit:
            params.update({'rows': limit})

        result_response = self.requests.get(url, params=params)
        err_desc = result_response.json().get('error')
        if err_desc:
            return {'err': 1, 'err_desc': err_desc}

        result = []
        for record in result_response.json().get('records'):
            data = {}
            for key, value in record.get('fields').items():
                data[key] = value
            data['id'] = record.get('recordid')
            template = Template(text_template)
            data['text'] = template.render(Context(data)).strip()
            result.append(data)

        return {'data': result}

    @endpoint(name='q',
              description=_('Query'),
              pattern=r'^(?P<query_slug>[\w:_-]+)/$',
              perm='can_access',
              show=False)
    def q(self, request, query_slug, **kwargs):
        query = get_object_or_404(Query, resource=self, slug=query_slug)
        return query.q(request, **kwargs)

    def create_query_url(self):
        return reverse('opendatasoft-query-new', kwargs={'slug': self.slug})


class Query(BaseQuery):
    resource = models.ForeignKey(
        to=OpenDataSoft,
        related_name='queries',
        verbose_name=_('Resource'))
    dataset = models.CharField(
        _('Dataset'),
        max_length=128, blank=False,
        help_text=_('dataset to query'),
    )
    text_template = models.TextField(
        verbose_name=_('Text template'),
        help_text=_(
            "Use Django's template syntax. Attributes can be accessed through {{ attributes.name }}"
        ),
        validators=[validate_template],
        blank=True
    )

    delete_view = 'opendatasoft-query-delete'
    edit_view = 'opendatasoft-query-edit'

    def q(self, request, **kwargs):
        return self.resource.search(
            request, dataset=self.dataset, text_template=self.text_template, **kwargs)

    def as_endpoint(self):
        endpoint = super(Query, self).as_endpoint(path=self.resource.q.endpoint_info.name)

        search_endpoint = self.resource.search.endpoint_info
        endpoint.func = search_endpoint.func
        endpoint.show_undocumented_params = False

        # Copy generic params descriptions from original endpoint
        # if they are not overloaded by the query
        for param in search_endpoint.parameters:
            if param in ('dataset', 'text_template') and getattr(self, param):
                continue
            endpoint.parameters[param] = search_endpoint.parameters[param]
        return endpoint
