# passerelle - uniform access to multiple data sources and services
# Copyright (C) 2020  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import re

from django.db import models
from django.utils import six
from django.utils.translation import ugettext_lazy as _

from passerelle.base.models import BaseResource
from passerelle.compat import json_loads
from passerelle.utils.api import endpoint
from passerelle.utils.jsonresponse import APIError


class SMSResource(BaseResource):
    category = _('SMS Providers')
    documentation_url = 'https://doc-publik.entrouvert.com/admin-fonctionnel/les-tutos/configuration-envoi-sms/'

    _can_send_messages_description = _('Sending messages is limited to the following API users:')

    default_country_code = models.CharField(verbose_name=_('Default country code'), max_length=3,
                                            default=u'33')
    default_trunk_prefix = models.CharField(verbose_name=_('Default trunk prefix'), max_length=2,
                                            default=u'0')  # Yeah France first !
    # FIXME: add regexp field, to check destination and from format
    max_message_length = models.IntegerField(_('Maximum message length'), default=160)

    def clean_numbers(self, destinations):
        numbers = []
        for dest in destinations:
            # most gateways needs the number prefixed by the country code, this is
            # really unfortunate.
            dest = dest.strip()
            number = ''.join(re.findall('[0-9]', dest))
            if dest.startswith('+'):
                number = '00' + number
            elif number.startswith('00'):
                # assumes 00 is international access code, remove it
                pass
            elif number.startswith(self.default_trunk_prefix):
                number = '00' + self.default_country_code + number[len(self.default_trunk_prefix):]
            else:
                raise APIError('phone number %r is unsupported (no international prefix, '
                               'no local trunk prefix)' % number)
            numbers.append(number)
        return numbers

    @endpoint(perm='can_send_messages', methods=['post'])
    def send(self, request, *args, **kwargs):
        try:
            data = json_loads(request.body)
            assert isinstance(data, dict), 'JSON payload is not a dict'
            assert 'message' in data, 'missing "message" in JSON payload'
            assert 'from' in data, 'missing "from" in JSON payload'
            assert 'to' in data, 'missing "to" in JSON payload'
            assert isinstance(data['message'], six.text_type), 'message is not a string'
            assert isinstance(data['from'], six.text_type), 'from is not a string'
            assert all(map(lambda x: isinstance(x, six.text_type), data['to'])), \
                'to is not a list of strings'
        except (ValueError, AssertionError) as e:
            raise APIError('Payload error: %s' % e)
        data['message'] = data['message'][:self.max_message_length]
        data['to'] = self.clean_numbers(data['to'])
        logging.info('sending SMS to %r from %r', data['to'], data['from'])
        stop = not bool('nostop' in request.GET)
        self.add_job('send_job',
                     text=data['message'], sender=data['from'], destinations=data['to'],
                     stop=stop)
        return {'err': 0}

    def send_job(self, *args, **kwargs):
        self.send_msg(**kwargs)
        SMSLog.objects.create(appname=self.get_connector_slug(), slug=self.slug)

    class Meta:
        abstract = True


@six.python_2_unicode_compatible
class SMSLog(models.Model):
    timestamp = models.DateTimeField(auto_now_add=True)
    appname = models.CharField(max_length=128, verbose_name='appname', null=True)
    slug = models.CharField(max_length=128, verbose_name='slug', null=True)

    def __str__(self):
        return '%s %s %s' % (self.timestamp, self.appname, self.slug)
